using System.Collections.ObjectModel;
using System.Text.Json;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Properties;
using View_by_Distance.Shared.Models.Stateless.Methods;

namespace View_by_Distance.Container.Models.Stateless.Methods;

internal abstract class Container
{

    private record FilePair(bool IsUnique,
                            List<string> Collection,
                            FilePath FilePath,
                            Item Item);

    internal static ReadOnlyCollection<Item> GetValidImageItems(IPropertyConfiguration propertyConfiguration, Models.Container container)
    {
        List<Item> results = [];
        foreach (Item item in container.Items)
        {
            if (!item.IsValidImageFormatExtension || propertyConfiguration.IgnoreExtensions.Contains(item.FilePath.ExtensionLowered))
                continue;
            results.Add(item);
        }
        return container.Items.Count == results.Count ? container.Items : results.AsReadOnly();
    }

    internal static DateTime[] GetContainerDateTimes(ReadOnlyCollection<Item> items)
    {
        DateTime[] results;
        long containerMinimumTicks = (from l in items select l.FilePath.LastWriteTicks).Min();
        long containerMaximumTicks = (from l in items select l.FilePath.LastWriteTicks).Max();
        results = [new(containerMinimumTicks), new(containerMaximumTicks)];
        return results;
    }

    internal static List<int> GetFilteredDistinctIds(IPropertyConfiguration propertyConfiguration, ReadOnlyCollection<Models.Container> readOnlyContainers)
    {
        List<int> results = [];
        ReadOnlyCollection<Item> filteredItems;
        foreach (Models.Container container in readOnlyContainers)
        {
            if (container.Items.Count == 0)
                continue;
            filteredItems = GetValidImageItems(propertyConfiguration, container);
            if (filteredItems.Count == 0)
                continue;
            foreach (Item item in filteredItems)
            {
                if (item.Property?.Id is null || item.ResizedFileHolder is null)
                    continue;
                if (results.Contains(item.Property.Id.Value))
                    continue;
                results.Add(item.Property.Id.Value);
            }
        }
        return results;
    }

    internal static (int, Models.Container[]) GetContainers(IPropertyConfiguration propertyConfiguration, ReadOnlyDictionary<int, Identifier>? splatNineIdentifiers, string _)
    {
        int count;
        Models.Container[] results;
        const bool useCeilingAverage = true;
        const string fileSearchFilter = "*";
        IDlibDotNet dlibDotNet = GetDlibDotNet();
        const string directorySearchFilter = "*";
        ReadOnlyDictionary<int, ReadOnlyCollection<FilePath>>? keyValuePairs = null;
        ReadOnlyCollection<string[]> filesCollection = IDirectory.GetFilesCollection(propertyConfiguration.RootDirectory, directorySearchFilter, fileSearchFilter, useCeilingAverage);
        ReadOnlyCollection<ReadOnlyCollection<FilePath>> filePathsCollection = IDirectory.GetFilePathCollections(propertyConfiguration, filesCollection);
        (count, results) = GetContainers(dlibDotNet, propertyConfiguration, propertyConfiguration.RootDirectory, keyValuePairs, splatNineIdentifiers, filePathsCollection, directorySearchFilter);
        return (count, results);
    }

    private static IDlibDotNet GetDlibDotNet() =>
        throw new NotImplementedException();

    private static (int, Models.Container[]) GetContainers(IDlibDotNet dlibDotNet, IPropertyConfiguration propertyConfiguration, string filesCollectionDirectory, ReadOnlyDictionary<int, ReadOnlyCollection<FilePath>>? keyValuePairs, ReadOnlyDictionary<int, Identifier>? splatNineIdentifiers, ReadOnlyCollection<ReadOnlyCollection<FilePath>> filePathsCollection, string directorySearchFilter)
    {
        List<Models.Container> results = [];
        string directory;
        List<Item>? items;
        Models.Container container;
        List<string> directories = [];
        Dictionary<string, List<Item>> directoryToItems = [];
        foreach (ReadOnlyCollection<FilePath> filePaths in filePathsCollection)
        {
            if (filePaths.Count == 0)
                continue;
            directory = filePaths[0].DirectoryFullPath;
            if (directory is null)
                continue;
            if (!directories.Contains(directory))
                directories.Add(directory);
            if (!directoryToItems.TryGetValue(directory, out items))
            {
                directoryToItems.Add(directory, []);
                if (!directoryToItems.TryGetValue(directory, out items))
                    throw new Exception();
            }
        }
        (string aResultsFullGroupDirectory, _) = dlibDotNet.GetResultsFullGroupDirectories();
        string aPropertySingletonDirectory = Path.Combine(aResultsFullGroupDirectory, propertyConfiguration.ResultSingleton);
        if (!Directory.Exists(aPropertySingletonDirectory))
            _ = Directory.CreateDirectory(aPropertySingletonDirectory);
        List<FilePair> filePairs = GetFilePairs(dlibDotNet, propertyConfiguration, aPropertySingletonDirectory, filesCollectionDirectory, keyValuePairs, splatNineIdentifiers, filePathsCollection, directorySearchFilter);
        foreach (FilePair filePair in filePairs)
        {
            if (!directoryToItems.TryGetValue(filePair.FilePath.DirectoryFullPath, out items))
            {
                directoryToItems.Add(filePair.FilePath.DirectoryFullPath, []);
                if (!directoryToItems.TryGetValue(filePair.FilePath.DirectoryFullPath, out items))
                    throw new Exception();
            }
            items.Add(filePair.Item);
        }
        foreach (KeyValuePair<string, List<Item>> keyValuePair in directoryToItems)
        {
            if (keyValuePair.Value.Count == 0)
                continue;
            container = new(keyValuePair.Key, new(keyValuePair.Value));
            results.Add(container);
        }
        return (filePairs.Count, results.ToArray());
    }

    internal static ReadOnlyCollection<Models.Container> GetContainers(IDlibDotNet dlibDotNet, IPropertyConfiguration propertyConfiguration, string facesFileNameExtension, string facesHiddenFileNameExtension, string eDistanceContentDirectory, string filesCollectionDirectory, ReadOnlyDictionary<int, ReadOnlyCollection<FilePath>>? keyValuePairs, ReadOnlyDictionary<int, Identifier>? splatNineIdentifiers, ReadOnlyCollection<ReadOnlyCollection<FilePath>> filePathsCollection)
    {
        Models.Container[] results;
        const string directorySearchFilter = "*";
        (_, results) = GetContainers(dlibDotNet, propertyConfiguration, filesCollectionDirectory, keyValuePairs, splatNineIdentifiers, filePathsCollection, directorySearchFilter);
        if (keyValuePairs is not null)
            DoGetFilePairsForRemaining(dlibDotNet, propertyConfiguration, facesFileNameExtension, facesHiddenFileNameExtension, eDistanceContentDirectory, filePathsCollection, directorySearchFilter);
        return results.AsReadOnly();
    }

    private static void DoGetFilePairsForRemaining(IDlibDotNet dlibDotNet, IPropertyConfiguration propertyConfiguration, string facesFileNameExtension, string facesHiddenFileNameExtension, string eDistanceContentDirectory, ReadOnlyCollection<ReadOnlyCollection<FilePath>> filePathsCollection, string directorySearchFilter)
    {
        const string extension = ".json";
        (_, string bResultsFullGroupDirectory) = dlibDotNet.GetResultsFullGroupDirectories();
        IReadOnlyDictionary<string, List<string>> fileNamesToFiles = GetFilesKeyValuePairs(filePathsCollection);
        string bMetaSingletonDirectory = Path.Combine(bResultsFullGroupDirectory, propertyConfiguration.ResultSingleton);
        if (!Directory.Exists(bMetaSingletonDirectory))
            _ = Directory.CreateDirectory(bMetaSingletonDirectory);
        _ = GetFilePairs(propertyConfiguration, directorySearchFilter, extension, bMetaSingletonDirectory, filePathsCollection, fileNamesToFiles);
        (string cResultsFullGroupDirectory, _, string dResultsFullGroupDirectory, _) = dlibDotNet.GetResultsFullGroupDirectories("Original");
        string cResizeSingletonDirectory = Path.Combine(cResultsFullGroupDirectory, propertyConfiguration.ResultSingleton);
        if (!Directory.Exists(cResizeSingletonDirectory))
            _ = Directory.CreateDirectory(cResizeSingletonDirectory);
        _ = GetFilePairs(propertyConfiguration, directorySearchFilter, extension, cResizeSingletonDirectory, filePathsCollection, fileNamesToFiles);
        string dFaceCollectionDirectory = Path.Combine(dResultsFullGroupDirectory, propertyConfiguration.ResultCollection);
        if (!Directory.Exists(dFaceCollectionDirectory))
            _ = Directory.CreateDirectory(dFaceCollectionDirectory);
        _ = GetFilePairs(propertyConfiguration, directorySearchFilter, extension, dFaceCollectionDirectory, filePathsCollection, fileNamesToFiles);
        string dFaceContentDirectory = Path.Combine(dResultsFullGroupDirectory, propertyConfiguration.ResultContent);
        if (!Directory.Exists(dFaceContentDirectory))
            _ = Directory.CreateDirectory(dFaceContentDirectory);
        AnyMovedFace(facesFileNameExtension, facesHiddenFileNameExtension, fileNamesToFiles, dFaceContentDirectory);
        AnyMovedDistance(facesFileNameExtension, fileNamesToFiles, eDistanceContentDirectory);
    }

    private static IReadOnlyDictionary<string, List<string>> GetFilesKeyValuePairs(ReadOnlyCollection<string[]> filesCollection)
    {
        Dictionary<string, List<string>> results = [];
        string fileNameWithoutExtension;
        string fileNameWithoutExtensionSecond;
        string fileNameWithoutExtensionSecondMinusOne;
        List<string>? collection;
        foreach (string[] files in filesCollection)
        {
            foreach (string file in files)
            {
                fileNameWithoutExtension = Path.GetFileNameWithoutExtension(file);
                fileNameWithoutExtensionSecond = Path.GetFileNameWithoutExtension(fileNameWithoutExtension);
                if (string.IsNullOrEmpty(fileNameWithoutExtensionSecond) || fileNameWithoutExtensionSecond == fileNameWithoutExtension)
                    continue;
                fileNameWithoutExtensionSecondMinusOne = fileNameWithoutExtensionSecond[..^1];
                if (!results.TryGetValue(fileNameWithoutExtensionSecondMinusOne, out collection))
                {
                    results.Add(fileNameWithoutExtensionSecondMinusOne, []);
                    if (!results.TryGetValue(fileNameWithoutExtensionSecondMinusOne, out collection))
                        throw new Exception();
                }
                collection.Add(file);
            }
        }
        return results;
    }

    private static IReadOnlyDictionary<string, List<string>> GetFilesKeyValuePairs(ReadOnlyCollection<ReadOnlyCollection<FilePath>> filePathsCollection)
    {
        Dictionary<string, List<string>> results = [];
        List<string>? collection;
        string fileNameWithoutExtensionMinusOne;
        foreach (ReadOnlyCollection<FilePath> filePaths in filePathsCollection)
        {
            foreach (FilePath filePath in filePaths)
            {
                fileNameWithoutExtensionMinusOne = filePath.NameWithoutExtension[..^1];
                if (!results.TryGetValue(fileNameWithoutExtensionMinusOne, out collection))
                {
                    results.Add(fileNameWithoutExtensionMinusOne, []);
                    if (!results.TryGetValue(fileNameWithoutExtensionMinusOne, out collection))
                        throw new Exception();
                }
                collection.Add(filePath.FullName);
            }
        }
        return results;
    }

    private static List<Shared.Models.FilePair> GetFilePairs(IPropertyConfiguration propertyConfiguration, string directorySearchFilter, string extension, string jsonGroupDirectory, ReadOnlyCollection<ReadOnlyCollection<FilePath>> filePathsCollection, IReadOnlyDictionary<string, List<string>> fileNamesToFiles)
    {
        List<Shared.Models.FilePair>? results = null;
        int renamed;
        const bool useCeilingAverage = true;
        ReadOnlyCollection<string[]>? jsonFilesCollection = null;
        IReadOnlyDictionary<string, List<string>>? compareFileNamesToFiles = null;
        for (int i = 0; i < short.MaxValue; i++)
        {
            renamed = 0;
            jsonFilesCollection = IDirectory.GetFilesCollection(jsonGroupDirectory, directorySearchFilter, extension, useCeilingAverage);
            compareFileNamesToFiles = GetFilesKeyValuePairs(jsonFilesCollection);
            renamed += LookForAbandoned(jsonFilesCollection, fileNamesToFiles, extension);
            results = IDirectory.GetFiles(propertyConfiguration, filePathsCollection, fileNamesToFiles, compareFileNamesToFiles);
            renamed += IDirectory.MaybeMove(propertyConfiguration, results, jsonGroupDirectory, extension);
            if (renamed == 0)
                break;
            if (i > 10)
                throw new NotImplementedException();
        }
        if (results is null || jsonFilesCollection is null || compareFileNamesToFiles is null)
            throw new NullReferenceException(nameof(results));
        return results;
    }

    private static int LookForAbandoned(ReadOnlyCollection<string[]> jsonFilesCollection, IReadOnlyDictionary<string, List<string>> fileNamesToFiles, string extension)
    {
        int result;
        bool check;
        bool moved = false;
        List<string> renameCollection = [];
        foreach (string[] files in jsonFilesCollection)
        {
            if (files.Length == 0)
                continue;
            check = AnyMoved(fileNamesToFiles, extension, renameCollection, files);
            if (!moved && check)
                moved = true;
        }
        if (renameCollection.Count > 0)
            IDirectory.MoveFiles(renameCollection, "{}", "{abd}");
        result = renameCollection.Count;
        if (moved && result == 0)
            result = 1;
        return result;
    }

    private static bool AnyMoved(IReadOnlyDictionary<string, List<string>> fileNamesToFiles, string extension, List<string> renameCollection, string[] files)
    {
        bool result = false;
        string checkFile;
        string directory;
        string fileNameWith;
        string checkDirectory;
        string directoryName;
        List<string>? collection;
        string fileNameWithoutExtension;
        List<string> directoryNames = [];
        string fileNameWithoutExtensionSecond;
        string fileNameWithoutExtensionSecondMinusOne;
        foreach (string file in files)
        {
            if (!file.EndsWith(extension))
                throw new Exception();
            fileNameWithoutExtension = Path.GetFileNameWithoutExtension(file);
            fileNameWithoutExtensionSecond = Path.GetFileNameWithoutExtension(fileNameWithoutExtension);
            if (string.IsNullOrEmpty(fileNameWithoutExtensionSecond) || fileNameWithoutExtensionSecond == fileNameWithoutExtension)
                continue;
            fileNameWithoutExtensionSecondMinusOne = fileNameWithoutExtensionSecond[..^1];
            if (!fileNamesToFiles.TryGetValue(fileNameWithoutExtensionSecondMinusOne, out collection))
                renameCollection.Add(file);
            else
            {
                directoryNames.Clear();
                directoryName = Path.GetFileName(Path.GetDirectoryName(file)) ?? throw new Exception();
                foreach (string f in collection)
                    directoryNames.Add(Path.GetFileName(Path.GetDirectoryName(f)) ?? throw new Exception());
                if (directoryNames.Count == 0 || directoryNames.Distinct().Count() != 1)
                    continue;
                if (directoryName != directoryNames[0])
                {
                    directory = Path.GetDirectoryName(Path.GetDirectoryName(file)) ?? throw new Exception();
                    checkDirectory = Path.Combine(directory, directoryNames[0]);
                    if (!Directory.Exists(checkDirectory))
                        _ = Directory.CreateDirectory(checkDirectory);
                    fileNameWith = collection.Count > 1 ? Path.GetFileName(file) : $"{Path.GetFileName(collection[0])}{extension}";
                    checkFile = Path.Combine(checkDirectory, fileNameWith);
                    if (!result)
                        result = true;
                    if (!File.Exists(checkFile))
                        File.Move(file, checkFile);
                    else
                    {
                        if (new FileInfo(file).LastWriteTime > new FileInfo(checkFile).LastWriteTime)
                            File.Delete(file);
                        else
                            File.Move(file, checkFile, true);
                    }
                }
            }
        }
        return result;
    }

    private static void AnyMovedFace(string extension, string hiddenExtension, IReadOnlyDictionary<string, List<string>> fileNamesToFiles, string jsonGroupDirectory)
    {
        string directory;
        string checkFile;
        string directoryName;
        List<string> files = [];
        string[] fileNameSegments;
        List<string> directories = [];
        files.AddRange(Directory.GetFiles(jsonGroupDirectory, $"*{extension}", SearchOption.AllDirectories));
        files.AddRange(Directory.GetFiles(jsonGroupDirectory, $"*{hiddenExtension}", SearchOption.AllDirectories));
        foreach (string file in files)
        {
            directory = Path.GetDirectoryName(file) ?? throw new Exception();
            if (!directories.Contains(directory))
                directories.Add(directory);
            directoryName = Path.GetFileName(directory);
            fileNameSegments = Path.GetFileName(file).Split('.');
            if (fileNameSegments[0] != directoryName)
            {
                fileNameSegments[0] = string.Empty;
                checkFile = Path.Combine(directory, $"{directoryName}{string.Join('.', fileNameSegments)}");
                if (!File.Exists(checkFile))
                    File.Move(file, checkFile);
                else
                {
                    if (new FileInfo(file).LastWriteTime > new FileInfo(checkFile).LastWriteTime)
                        File.Delete(file);
                    else
                        File.Move(file, checkFile, true);
                }
            }
        }
        if (directories.Count > 0)
            AnyMovedFace(fileNamesToFiles, directories);
    }

    private static void AnyMovedFace(IReadOnlyDictionary<string, List<string>> fileNamesToFiles, List<string> directories)
    {
        bool result = false;
        string checkFile;
        string subDirectory;
        string directoryName;
        string checkDirectory;
        List<string>? collection;
        string directoryNameWith;
        string directoryNameMinusOne;
        List<string> directoryNames = [];
        foreach (string directory in directories)
        {
            directoryNameMinusOne = Path.GetFileName(directory)[..^1];
            if (!fileNamesToFiles.TryGetValue(directoryNameMinusOne, out collection))
                throw new Exception();
            directoryNames.Clear();
            foreach (string f in collection)
                directoryNames.Add(Path.GetFileName(Path.GetDirectoryName(f)) ?? throw new Exception());
            if (directoryNames.Count == 0 || directoryNames.Distinct().Count() != 1)
                continue;
            directoryName = Path.GetFileName(Path.GetDirectoryName(directory)) ?? throw new Exception();
            if (directoryName != directoryNames[0])
            {
                subDirectory = Path.GetDirectoryName(Path.GetDirectoryName(directory)) ?? throw new Exception();
                checkDirectory = Path.Combine(subDirectory, directoryNames[0]);
                if (!Directory.Exists(checkDirectory))
                    _ = Directory.CreateDirectory(checkDirectory);
                directoryNameWith = collection.Count > 1 ? directoryName : $"{Path.GetFileNameWithoutExtension(collection[0])}";
                checkFile = Path.Combine(checkDirectory, directoryNameWith);
                if (!result)
                    result = true;
                if (!Directory.Exists(checkFile))
                    Directory.Move(directory, checkFile);
                else
                {
                    if (new DirectoryInfo(directory).LastWriteTime > new DirectoryInfo(checkFile).LastWriteTime)
                        Directory.Delete(directory, recursive: true);
                    else
                    {
                        Directory.Delete(checkFile, recursive: true);
                        Directory.Move(directory, checkFile);
                    }
                }
            }
        }
    }

    private static void AnyMovedDistance(string extension, IReadOnlyDictionary<string, List<string>> fileNamesToFiles, string jsonGroupDirectory)
    {
        bool result = false;
        string fileName;
        string checkFile;
        string directory;
        List<string>? collection;
        string[] fileNameSegments;
        List<string> fileNames = [];
        string fileNameSegmentsZeroMinusOne;
        string[] files = Directory.GetFiles(jsonGroupDirectory, $"*{extension}", SearchOption.AllDirectories);
        foreach (string file in files)
        {
            fileName = Path.GetFileName(file);
            fileNameSegments = fileName.Split('.');
            fileNameSegmentsZeroMinusOne = fileNameSegments[0][..^1];
            if (!fileNamesToFiles.TryGetValue(fileNameSegmentsZeroMinusOne, out collection))
                continue;
            fileNames.Clear();
            foreach (string f in collection)
                fileNames.Add(Path.GetFileNameWithoutExtension(f) ?? throw new Exception());
            if (fileNames.Count == 0 || fileNames.Distinct().Count() != 1)
                continue;
            if (fileNameSegments[0] != fileNames[0])
            {
                fileNameSegments[0] = string.Empty;
                directory = Path.GetDirectoryName(file) ?? throw new Exception();
                checkFile = Path.Combine(directory, $"{fileNames[0]}{string.Join('.', fileNameSegments)}");
                if (!result)
                    result = true;
                if (!File.Exists(checkFile))
                    File.Move(file, checkFile);
                else
                {
                    if (new FileInfo(file).LastWriteTime > new FileInfo(checkFile).LastWriteTime)
                        File.Delete(file);
                    else
                        File.Move(file, checkFile, true);
                }
            }
        }
    }

    private static List<FilePair> GetFilePairs(IDlibDotNet dlibDotNet, IPropertyConfiguration propertyConfiguration, string aPropertySingletonDirectory, string filesCollectionDirectory, ReadOnlyDictionary<int, ReadOnlyCollection<FilePath>>? keyValuePairs, ReadOnlyDictionary<int, Identifier>? splatNineIdentifiers, ReadOnlyCollection<ReadOnlyCollection<FilePath>> filePathsCollection, string directorySearchFilter)
    {
        List<FilePair> results = [];
        const string extension = ".json";
        List<Shared.Models.FilePair> filePairs;
        string jsonGroupDirectory = aPropertySingletonDirectory;
        int maxDegreeOfParallelism = Environment.ProcessorCount;
        int filesCollectionDirectoryLength = filesCollectionDirectory.Length;
        ParallelOptions parallelOptions = new() { MaxDegreeOfParallelism = maxDegreeOfParallelism };
        IReadOnlyDictionary<string, List<string>> fileNamesToFiles = GetFilesKeyValuePairs(filePathsCollection);
        filePairs = GetFilePairs(propertyConfiguration, directorySearchFilter, extension, jsonGroupDirectory, filePathsCollection, fileNamesToFiles);
        _ = Parallel.For(0, filePairs.Count, parallelOptions, (i, state) =>
            ParallelFor(dlibDotNet, propertyConfiguration, jsonGroupDirectory, extension, keyValuePairs, splatNineIdentifiers, filesCollectionDirectoryLength, filePairs[i], results));
        return results;
    }

    private static void ParallelFor(IDlibDotNet dlibDotNet, IPropertyConfiguration propertyConfiguration, string jsonGroupDirectory, string extension, ReadOnlyDictionary<int, ReadOnlyCollection<FilePath>>? keyValuePairs, ReadOnlyDictionary<int, Identifier>? splatNineIdentifiers, int rootDirectoryLength, Shared.Models.FilePair filePair, List<FilePair> results)
    {
        dlibDotNet?.Tick();
        bool abandoned = false;
        FileHolder sourceDirectoryFileHolder;
        Property? property = GetProperty(filePair);
        FileHolder fileHolder = IFileHolder.Get(filePair.Path);
        ExifDirectory? exifDirectory = GetExifDirectory(filePair);
        FilePath filePath = FilePath.Get(propertyConfiguration, fileHolder, index: null);
        bool? fileSizeChanged = property is not null ? property.FileSize != filePath.Length : null;
        bool isValidImageFormatExtension = propertyConfiguration.ValidImageFormatExtensions.Contains(filePath.ExtensionLowered);
        bool? shouldIgnore = property is null || property.Keywords is null ? null : propertyConfiguration.IgnoreRulesKeyWords.Any(l => property.Keywords.Contains(l));
        bool? isArchive = filePath.Id is null || splatNineIdentifiers is null ? null : splatNineIdentifiers.TryGetValue(filePath.Id.Value, out Identifier? identifier);
        if (property is not null && filePath.Id is not null && filePath.HasIgnoreKeyword is not null && filePath.HasDateTimeOriginal is not null)
        {
            char? change;
            ReadOnlyCollection<FilePath>? filePaths = null;
            char hasIgnoreKeyword = IId.GetHasIgnoreKeyword(filePath).ToString()[0];
            char hasDateTimeOriginal = IId.GetHasDateTimeOriginal(propertyConfiguration, filePath).ToString()[0];
            char missingDateTimeOriginal = IId.GetMissingDateTimeOriginal(propertyConfiguration, filePath).ToString()[0];
            if (shouldIgnore is not null && shouldIgnore.Value)
            {
                if (filePath.NameWithoutExtension[^1] == hasIgnoreKeyword)
                    change = null;
                else
                {
                    change = hasIgnoreKeyword;
                    if (keyValuePairs is null || !keyValuePairs.TryGetValue(filePath.Id.Value, out filePaths) || filePaths is null)
                        throw new NotSupportedException($"Rename File! <{filePath.FileNameFirstSegment}>");
                }
            }
            else if ((shouldIgnore is null || !shouldIgnore.Value) && property.DateTimeOriginal is null)
            {
                if (filePath.NameWithoutExtension[^1] == missingDateTimeOriginal)
                    change = null;
                else
                {
                    change = missingDateTimeOriginal;
                    if (keyValuePairs is null || !keyValuePairs.TryGetValue(filePath.Id.Value, out filePaths) || filePaths is null)
                        throw new NotSupportedException($"Rename File! <{filePath.FileNameFirstSegment}>");
                }
            }
            else if (filePath.NameWithoutExtension[^1] != hasDateTimeOriginal)
            {
                change = hasDateTimeOriginal;
                if (keyValuePairs is null || !keyValuePairs.TryGetValue(filePath.Id.Value, out filePaths) || filePaths is null)
                    throw new NotSupportedException($"Rename File! <{filePath.FileNameFirstSegment}>");
            }
            else
                change = null;
            if (filePaths is not null && change is not null)
                RenameFile(extension, filePair, filePath, change.Value, filePaths);
        }
        string relativePath = Shared.Models.Stateless.Methods.IPath.GetRelativePath(filePair.Path, rootDirectoryLength, forceExtensionToLower: true);
        bool? lastWriteTimeChanged = property is not null ? propertyConfiguration.PropertiesChangedForProperty || property.LastWriteTime.Ticks != filePath.LastWriteTicks : null;
        if (filePair.Match is not null)
            sourceDirectoryFileHolder = IFileHolder.Get(filePair.Match);
        else if (!filePair.IsUnique)
            sourceDirectoryFileHolder = IFileHolder.Get(Path.GetFullPath(string.Concat(jsonGroupDirectory, relativePath, extension)));
        else
        {
            string fileName = Path.GetFileName(filePair.Path);
            CombinedEnumAndIndex cei = Shared.Models.Stateless.Methods.IPath.GetCombinedEnumAndIndex(propertyConfiguration, filePath);
            string directory = Path.Combine(jsonGroupDirectory, cei.Combined);
            string jsonFileName = $"{fileName}{extension}";
            string fullFileName = Path.Combine(directory, jsonFileName);
            MoveIf(jsonFileName, cei, directory, fullFileName);
            sourceDirectoryFileHolder = IFileHolder.Get(fullFileName);
        }
        if (sourceDirectoryFileHolder.CreationTime is not null && sourceDirectoryFileHolder.LastWriteTime is not null && filePath.LastWriteTicks != sourceDirectoryFileHolder.CreationTime.Value.Ticks)
        {
            File.SetCreationTime(sourceDirectoryFileHolder.FullName, new(filePath.LastWriteTicks));
            File.SetLastWriteTime(sourceDirectoryFileHolder.FullName, sourceDirectoryFileHolder.LastWriteTime.Value);
        }
        Item item = Item.Get(filePath, sourceDirectoryFileHolder, relativePath, isArchive, filePair.IsNotUniqueAndNeedsReview, filePair.IsUnique, isValidImageFormatExtension, property, abandoned, fileSizeChanged, lastWriteTimeChanged);
        lock (results)
            results.Add(new(filePair.IsUnique, filePair.Collection, filePath, item));
    }

    private static Property? GetProperty(Shared.Models.FilePair filePair)
    {
        Property? result;
        if (filePair.Match is null)
            result = null;
        else
        {
            string json = File.ReadAllText(filePair.Match);
            if (string.IsNullOrEmpty(json))
                result = null;
            else
                result = JsonSerializer.Deserialize(json, PropertyGenerationContext.Default.Property);
        }
        return result;
    }

    private static ExifDirectory? GetExifDirectory(Shared.Models.FilePair filePair)
    {
        ExifDirectory? result;
        if (filePair.Match is null)
            result = null;
        else
        {
            string json = File.ReadAllText(filePair.Match);
            if (string.IsNullOrEmpty(json))
                result = null;
            else
                result = JsonSerializer.Deserialize(json, ExifDirectorySourceGenerationContext.Default.ExifDirectory);
        }
        return result;
    }

    private static void RenameFile(string extension, Shared.Models.FilePair filePair, FilePath filePath, char change, ReadOnlyCollection<FilePath> filePaths)
    {
        string checkFile;
        if (filePath.DirectoryFullPath.Contains("Results") && filePath.DirectoryFullPath.Contains("Resize"))
            File.Delete(filePath.FullName);
        if (!string.IsNullOrEmpty(filePair.Match))
        {
            string directory = Path.GetDirectoryName(filePair.Match) ?? throw new Exception();
            string fileNameWithoutExtension = Path.GetFileNameWithoutExtension(filePair.Match);
            string fileNameWithoutExtensionSecond = Path.GetFileNameWithoutExtension(Path.GetFileNameWithoutExtension(filePair.Match));
            string extensionSecond = Path.GetExtension(fileNameWithoutExtension);
            checkFile = Path.Combine(directory, $"{fileNameWithoutExtensionSecond[..^1]}{change}{extensionSecond}{extension}");
            if (!File.Exists(checkFile))
                File.Move(filePair.Match, checkFile);
        }
        foreach (FilePath f in filePaths)
        {
            checkFile = Path.Combine(f.DirectoryFullPath, $"{f.NameWithoutExtension[..^1]}{change}{f.ExtensionLowered}");
            if (File.Exists(checkFile))
                continue;
            File.Move(f.FullName, checkFile);
        }
    }

    private static void MoveIf(string fileName, CombinedEnumAndIndex cei, string directory, string fullFileName)
    {
        string[] segments = directory.Split(cei.Combined);
        string? checkDirectory = segments.Length == 1 ?
            Path.Combine(segments[0], $"{cei.Combined[2..]}") :
                segments.Length == 2 ?
                    $"{segments[0]}{cei.Combined[2..]}{segments[1]}" :
                        null;
        if (checkDirectory is not null && Directory.Exists(checkDirectory))
        {
            string checkFile = Path.Combine(checkDirectory, fileName);
            if (File.Exists(checkFile))
                File.Move(checkFile, fullFileName);
        }
    }

    internal static List<string> GetFilteredDistinctFileNameFirstSegments(IPropertyConfiguration propertyConfiguration, ReadOnlyCollection<Models.Container> readOnlyContainers)
    {
        List<string> results = [];
        ReadOnlyCollection<Item> filteredItems;
        foreach (Models.Container container in readOnlyContainers)
        {
            if (container.Items.Count == 0)
                continue;
            filteredItems = GetValidImageItems(propertyConfiguration, container);
            if (filteredItems.Count == 0)
                continue;
            foreach (Item item in filteredItems)
            {
                if (item.Property?.Id is null || item.ResizedFileHolder is null)
                    continue;
                if (results.Contains(item.FilePath.FileNameFirstSegment))
                    continue;
                results.Add(item.FilePath.FileNameFirstSegment);
            }
        }
        return results;
    }

    internal static ReadOnlyCollection<Item> GetValidImageItems(IPropertyConfiguration propertyConfiguration, ReadOnlyCollection<Models.Container> containers, bool distinctItems, bool filterItems)
    {
        List<Item> results = [];
        List<int> distinct = [];
        ReadOnlyCollection<Item> filteredItems;
        foreach (Models.Container container in containers)
        {
            if (container.Items.Count == 0)
                continue;
            if (!filterItems)
                filteredItems = container.Items;
            else
            {
                filteredItems = GetValidImageItems(propertyConfiguration, container);
                if (filteredItems.Count == 0)
                    continue;
            }
            foreach (Item item in filteredItems)
            {
                if (item.Property?.Id is null || item.ResizedFileHolder is null)
                    continue;
                if (distinctItems)
                {
                    if (distinct.Contains(item.Property.Id.Value))
                        continue;
                    distinct.Add(item.Property.Id.Value);
                }
                results.Add(item);
            }
        }
        return results.AsReadOnly();
    }

}